

module RefacMoveDef(Direction(..),move) where

import PrettyPrint
import PosSyntax
import Maybe
import TypedIds
import UniqueNames
import List 
import PNT 

import RefacUtils
import RefacRmDef

data Direction = UptoTopLevel | UpOneLevel | Down

{--------This function handles refactorings involving moving a defintion--------
 According to the Haskell's  syntax, a declaration may occur in one of the following six contexts:
  1. A top level declaration in the module:
            HsModule SrcLoc ModuleName (Maybe [HsExportSpecI i]) [HsImportDeclI i] ds
  2. A local declaration in a Match:
            HsMatch SrcLoc i [p] (HsRhs e) ds
  3. A local declaration in a pattern binding:
            HsPatBind SrcLoc p (HsRhs e) ds
  4. A local declaration in a Let expression:
            HsLet ds e
  5. A local declaration in a Case alternative:
            HsAlt SrcLoc p (HsRhs e) ds
  6. A local declaration in a Do statement:
            HsLetStmt ds (HsStmt e p ds)           
-}
move direction args
  = do let fileName = ghead "filename"  args 
           --fileName'= moduleName fileName
           --modName = Module fileName'  
           row = read (args!!1)::Int
           col = read (args!!2)::Int
       modName <- moduleName fileName
       (inscps, _, mod, tokList) <- parseSourceFile fileName  
       let pn = locToPName fileName row col mod                                         
       if pn /= defaultPN
         then 
          case direction  of
               UptoTopLevel ->liftToTopLevel modName fileName (inscps, mod, tokList) pn
               UpOneLevel   ->liftOneLevel   modName fileName (inscps, mod, tokList) pn   
               Down         ->demote         modName fileName (mod, tokList)  pn
          else error "\nInvalid cursor position!\n"
 
{- Refactoring Names: 'liftToTopLevel'
   This refactoring lifts a local function/pattern binding to the top level of the module, so as to 
    make it accessible to  other functions in the current module, and those modules that import 
    current module.  
    
   In the current implementation, a definition will be lifted only if none of the identifiers defined in this
   definition will cause name clash/capture problems in the current module after lifting. 

   In the case that the whole current module is exported implicitly, the lifted identifier will be  exported
   automatically after lifting. If the identifier will cause name clash/ambiguous occurrence problem in a 
   client module, it will be hided in the import declaration of the client module (Note: this might not be 
   the best solution, we prefer hiding it in the server module instead of in the client module in the final version).

   In the case of indirect importing, it might be time-consuming to trace whether the lifted identifier
   will cause any problem in a client module that indirectly imports the current  module. The current solution is:
   suppose a defintion is lifted to top level in module A, and module A is imported and exported by module B, then
   the lifted identifier will be hided in the import declaration of B no matter whether it causes problems in 
   module B or not.  
 
   Function name: liftToTopLevel 
   parameters: fileName--current file name. 
               mod -- the scoped abstract syntax tree of the module.
               pn  -- the function/pattern name to be lifted.        
-}

liftToTopLevel modName fileName (inscps, mod, tokList) pn   
  = if isLocalFunOrPatName pn mod     
      then do ((mod',declPns),((tokList',m),_))<-runStateT liftToMod ((tokList,unmodified),(fileName, initValue))            
              if wholeModIsExported mod  
               then do clients<-clientModsAndFiles modName  
                       refactoredClients <- mapM (liftingInClientMod modName declPns) clients
                       writeRefactoredFiles $ ((fileName,m),(tokList',mod')):refactoredClients                          
               else do writeRefactoredFiles [((fileName,m), (tokList',mod'))]
      else error "\nThe identifier is not a local function/pattern name!" 
   
     where 
          {-step1: divide the module's top level declaration list into three parts:
            'parant' is the top level declaration containing the lifted declaration,
            'before' and `after` are those declarations before and after 'parent'.
            step2: get the declarations to be lifted from parent, bind it to liftedDecls 
            step3: remove the lifted declarations from parent and extra arguments may be introduce.
            step4. test whether there are any names need to be renamed. 
          -}       
       liftToMod = do let (before, parent,after)=divideDecls (declsIn mod) pn                                 
                      when (isClassDecl $ ghead "liftToMod" parent) 
                            $ error "Sorry, the refactorer cannot lift a definition from a class declaration!"
                      when (isInstDecl $ ghead "liftToMod" parent)
                            $ error "Sorry, the refactorer cannot lift a definition from a instance declaration!"           
                      let liftedDecls=definingDecls [pn] (allDeclsIn parent) True
                          declaredPns=nub $ concatMap definedNames liftedDecls                                         
                      pns<-pnsNeedRenaming inscps mod parent liftedDecls declaredPns
                      (_,dd)<-hsFreeAndDeclared mod
                      if pns==[] 
                        then do (parent',liftedDecls',paramAdded)<-addParamsToParentAndLiftedDecl pn dd parent liftedDecls
                                let liftedDecls''=if paramAdded then filter isFunOrPatBind liftedDecls'
                                                                else liftedDecls'
                                --True means the new decl will be at the same level with its parant. 
                                parent'<-moveDecl declaredPns parent  Nothing True paramAdded parent' 
                                return (replaceDecls mod (before++parent'++liftedDecls''++after), declaredPns)
                        else askRenamingMsg pns "lifting"


--get all the declarations define in the scope of t
allDeclsIn t = fromMaybe [] (applyTU (full_tdTU (constTU [] `adhocTU` decl)) t)               
               where decl (d::HsDeclP)
                       |isFunBind d || isPatBind d || isTypeSig d = Just [d]
                     decl _ = Just [] 

askRenamingMsg pns str 
  = error ("The identifier(s):" ++ showPns pNtoQuotedName pns ++
           " will cause name clash/capture or ambiguity occurrence problem after "
           ++ str ++", please do renaming first!")             

--Get the subset of 'pns' that need to be renamed before lifting.
pnsNeedRenaming inscps dest parent liftedDecls pns
   =do r<-mapM pnsNeedRenaming' pns   
       return (concat r)
  where                 
     pnsNeedRenaming' pn
       = do (f,d)<-freeAndDeclaredVbls dest  --f: free variable names that may be shadowed by pn
                                             --d: declaread variables names that may clash with pn
            vs<-hsVisibleVbls pn parent      --vs: declarad varaibles that may shadow pn          
            let inscpNames = inScopeNames [inscps]
                vars = map pNtoName (nub (f `union` d `union` vs) \\ [pn]) -- `union` inscpNames            
            if elem (pNtoName pn) vars  || isInScopeAndUnqualified (pNtoName pn) inscps && findEntity pn dest 
               then return [pn]
               else return []
     --This pNtoName takes into account the qualifier.
     pNtoName (PN (UnQual i) orig)=i    
     pNtoName (PN (Qual (Module modName) i ) orig)=modName ++ "." ++ i        
 
--can not simply use PNameToExp, PNameToPat here because of the location information. 
addParamsToParent pn [] t = return t
addParamsToParent pn params t
   =applyTP(full_buTP (idTP `adhocTP` inExp 
                            `adhocTP` inDecl    
                            `adhocTP` inMatch)) t 
   where              
          inExp (exp@(Exp (HsId (HsVar (PNT pname ty loc))))::HsExpP)
            | pname==pn
             = do  addParamsInToks pn paramNames exp True
                   exps'<-mapM pNameToExp params
                   return (foldl addParamToExp exp exps')

          inExp x =return x
         
          inMatch ((HsMatch loc  fun  pats rhs  decls)::HsMatchP)
            |pNTtoPN fun==pn
             = do addParamsInToks pn paramNames fun False     
                  --idToPat is used here because the this does not afftect the AST (only affect token stream)
                  return (HsMatch loc  fun  ((map simpleIdToPat paramNames)++pats)  rhs  decls)

          inMatch x=return x

          inDecl ((Dec (HsPatBind loc p rhs ds))::HsDeclP)
            | patToPName p==pn
             =do addParamsInToks pn paramNames p False
                 --simpleIdToPat is used here because the this does not afftect the AST (only affect token stream)
                 return (Dec (HsFunBind loc [HsMatch loc (patToPNT p) (map simpleIdToPat paramNames) rhs ds]))
          inDecl x=return x  
           
          simpleIdToPat :: String -> HsPatP
          simpleIdToPat name
             =Pat (HsPId (HsVar (PNT (PN (UnQual name) (S loc0)) Value (N (Just loc0)))))  

          addParamToExp  exp param
              =(Exp (HsApp exp param)) 

          paramNames=map pNtoName params

          addParamsInToks  pn  paramNames t isExp
            =do ((toks,_),others)<-get        --This should be refactored
                let (startPos,endPos)=startEndLoc toks  t  --get the scope of the t
                     --compose a new token. (illegal according to the lexer)
                    newToken=if isExp 
                               then --actually sometimes brackets are not necessay.
                                    (Varid,(startPos,"("++ foldl (\x y->x++" "++y) (pNtoName pn) paramNames++")")) 
                               else (Varid,(startPos, foldl (\x y->x++" "++y) (pNtoName pn) paramNames))
                    --replace the token containing pn by newTtoken --not quited wheter it should just one token replaced.
                    toks'=replaceToks toks startPos endPos [newToken] 
                put ((toks',modified),others)

--Do refactoring in the client module.
-- that is to hide the identifer in the import declaration if it will cause any problem in the client module.
liftingInClientMod serverModName pns (modName, fileName)
  = do (inscps, exps ,mod ,ts) <- parseSourceFile fileName
       let modNames = willBeUnQualImportedBy serverModName mod   
       if isJust modNames
        then let pns' = namesNeedToBeHided mod exps (fromJust modNames) pns
             in if pns' /= [] 
                 then do (mod', ((ts',m),_))<-runStateT (addHiding serverModName mod pns') ((ts,unmodified),(fileName, initValue))
                         return ((fileName,m), (ts',mod')) 
                 else return ((fileName,unmodified), (ts,mod))
        else return ((fileName,unmodified),(ts,mod))


--Test whether an identifier defined in the modules specified by 'names' will be exported by current module.
willBeExportedByClientMod names mod
  = let exps = hsModExports mod
    in if isNothing exps 
          then False
          else any isJust $ map (\y-> (find (\x-> (x==ModuleE y)) (fromJust exps))) names

--get the module name or alias name by which the lifted identifier will be imported automatically.
willBeUnQualImportedBy modName mod
   = let imps  = hsModImports mod
         ms =filter (\(HsImportDecl _ modName1 qualify  as h)->modName==modName1 && (not qualify) && 
                          (isNothing h || (isJust h && ((fst (fromJust h))==True)))) imps
         in if ms==[] then Nothing
                      else Just $ nub $ map getModName ms

         where getModName (HsImportDecl _ modName qualify  as h)
                 = if isJust as then fromJust as
                                else modName

--get the subset of 'pns', which need to be hided in the import declaration in module 'mod'
namesNeedToBeHided mod exps modNames  pns
  = if willBeExportedByClientMod modNames mod
      then pns 
      else concatMap needToBeHided pns
    where
      needToBeHided  pn 
        = let name = pNtoName pn 
          in if (findNameInUseLoc name (hsModDecls mod) --the same name is used in the module unqualifiedly
                || findUnqualEntInExport name (hsModExports mod)  --the same name is exported unqualifiedly by an Ent decl
                || causeNameClashInClientExports name mod exps)
              then [pn] 
              else []


-- **************************************************************************************************************--

{- Refactoring Names: 'liftOneLevel'
   Descritption:
    this refactoring lifts a local function/pattern binding only one level up. 
    By 'lifting one-level up' ,I mean: 
    case1: In a module (HsModule SrcLoc ModuleName (Maybe [HsExportSpecI i]) [HsImportDeclI i] ds):
           A local declaration D  will be lifted to the same level as the 'ds', if D is in the 
           where clause of one of ds's element declaration.

    case2: In a match ( HsMatch SrcLoc i [p] (HsRhs e) ds) :
          A local declaration D  will be lifted to the same level as the 'ds', if D is in the 
           where clause of one of ds's element declaration.
           A declaration D,say,in the rhs expression 'e' will be lifted to 'ds' if D is Not local to
           other declaration list in 'e'

    case3: In a pattern  binding (HsPatBind SrcLoc p (HsRhs e) ds):
           A local declaration D  will be lifted to the same level as the 'ds', if D is in the 
           where clause of one of ds's element declaration.
           A declaration D,say,in the rhs expression 'e' will be lifted to 'ds' if D is Not local to
           other declaration list in 'e'

    case4: In the Lex expression (Exp (HsLet ds e):
           A local declaration D  will be lifted to the same level as the 'ds', if D is in the 
           where clause of one of ds's element declaration.
           A declaration D, say, in the expression 'e' will be lifted to 'ds' if D is not local to
           other declaration list in 'e'
    case5: In the case Alternative expression:(HsAlt loc p rhs ds)
           A local declaration D  will be lifted to the same level as the 'ds', if D is in the 
           where clause of one of ds's element declaration.
           A declaration D in 'rhs' will be lifted to 'ds' if D is not local to other declaration 
           list in 'rhs'.

    case6: In the do statement expression:(HsLetStmt ds stmts)
           A local declaration D  will be lifted to the same level as the 'ds', if D is in the 
           where clause of one of ds's element declaration.
           A declaration D in 'stmts' will be lifted to 'ds' if D is not local to other declaration 
           list in 'stmts'.           
    
Function name: liftOneLevel 
parameters: fileName--current file name.
            mod -- the scoped abstract syntax tree of the module.
            pn  -- the function/pattern name to be lifted.
       
-}

liftOneLevel modName fileName (inscps, mod, tokList) pn   
   = if isLocalFunOrPatName pn mod
        then do (mod', ((tokList',m),_))<-liftOneLevel' 
                let (b, pns) = liftedToTopLevel pn mod
                if b &&  wholeModIsExported mod  
                  then do clients<-clientModsAndFiles modName  
                          refactoredClients <- mapM (liftingInClientMod modName pns) clients
                          writeRefactoredFiles $ ((fileName,m),(tokList',mod')):refactoredClients     
                  else writeRefactoredFiles [((fileName,m), (tokList',mod'))]
        else error "\nThe identifer is not a function/pattern name!"

   where
      liftOneLevel'=runStateT (applyTP ((once_tdTP (failTP `adhocTP` liftToMod
                                                           `adhocTP` liftToMatch
                                                           `adhocTP` liftToPattern 
                                                           `adhocTP` liftToLet
                                                           `adhocTP` liftToAlt
                                                           `adhocTP` liftToLetStmt))
                                          `choiceTP` failure) mod) ((tokList,unmodified),(fileName,initValue))
           where          
             --1. The defintion will be lifted to top level
             liftToMod (mod@(HsModule loc name exps imps ds):: HsModuleP)  
                | findDecl pn (declsIn ds) False --False means not taking type signature into account 
                  =do ds'<-worker mod ds pn 
                      return (HsModule loc name exps imps ds')
             liftToMod  _ =mzero
         
             --2. The definition will be lifted to the declaration list of a match
             liftToMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
                 | findDecl pn (declsIn ds) False
                  =do ds'<-worker match ds pn            
                      return (HsMatch loc1 name pats rhs ds')

             liftToMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
                 | findDecl pn (declsIn rhs) False
                  = do let destPos=if ds/=[] then Just (fst (startEndLoc tokList (ghead "liftToMatch" ds)))
                                             else Nothing
                       --False mean the new decl will be a local decl of it's parent.
                       (rhs',liftedDecls)<-doLifting match rhs pn destPos False  
                       return (HsMatch loc1 name pats rhs' (liftedDecls++ds)) 
                
             liftToMatch _ =mzero

             --3. The definition will be lifted to the declaration list of a pattern binding 
             liftToPattern (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                | findDecl pn (declsIn ds) False
                  =do ds'<-worker pat ds pn 
                      return (Dec (HsPatBind loc p rhs ds'))

             liftToPattern (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                | findDecl pn  (declsIn rhs) False
                  =do let  destPos=if ds/=[] then Just (fst (startEndLoc tokList (ghead "liftToPattern" ds))) --no problem with 'head'
                                             else Nothing
                      (rhs',liftedDecls)<-doLifting pat rhs pn destPos False
                      return (Dec (HsPatBind loc p rhs' (liftedDecls++ds)))
             liftToPattern _=mzero

             --4. The definition will be lifted to the declaration list in a let expresiion.
             liftToLet (letExp@(Exp (HsLet ds e))::HsExpP)
               | findDecl pn (declsIn ds) False
                =do ds' <-worker letExp ds pn  
                    return (Exp (HsLet ds' e))

             liftToLet (letExp@(Exp (HsLet ds e))::HsExpP)  --Attention: ds can be empty!
               | findDecl pn (declsIn e) False
                =do let destPos=if ds/=[] then Just (fst (startEndLoc tokList (ghead "liftToLet" ds)))  --no problem with 'head'
                                          else let (startPos,endPos)=startEndLoc tokList letExp
                                                   expToks= getToks (startPos,endPos) tokList 
                                               in Just ((tokenPos.(ghead "liftToLet:else")) $ dropWhile (not.isIn) expToks)
                    (e',liftedDecls)<-doLifting letExp e pn destPos False
                    return (Exp (HsLet (liftedDecls++ds) e')) 
             liftToLet _ =mzero
 
           
             --5. The definition will be lifted to the declaration list in a alt
             liftToAlt (alt@(HsAlt loc p rhs ds)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
                |findDecl pn (declsIn ds) False
                =do ds'<-worker alt ds pn 
                    return (HsAlt loc p rhs ds')

             liftToAlt (alt@(HsAlt loc p rhs ds)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
                |findDecl pn (declsIn rhs) False
                =do let  destPos=if ds/=[] then Just (fst (startEndLoc tokList (ghead "liftToAlt" ds)))
                                           else Nothing
                    (rhs',liftedDecls)<-doLifting  alt rhs pn destPos False
                    return (HsAlt loc p rhs' (liftedDecls++ds))
             liftToAlt _=mzero

             --6. The defintion will be lifted to the declaration list in a let statement.
             liftToLetStmt (letStmt@(HsLetStmt ds stmts):: (HsStmt (HsExpP) (HsPatP) [HsDeclP]))
                |findDecl pn (declsIn ds) False
               =do ds'<-worker letStmt ds pn  
                   return (HsLetStmt ds' stmts)
              
             liftToLetStmt (letStmt@(HsLetStmt ds stmts):: (HsStmt (HsExpP) (HsPatP) [HsDeclP])) --attention: Let can be empty!
                |findDecl pn (declsIn stmts) False
               =do let destPos=if ds/=[] then Just (fst (startEndLoc tokList (ghead "liftToLetStmt:then" ds)))
                                         else error "lifting a declaration to a empty 'let clause' has not been implmentation!"
                   (stmts',liftedDecls)<-doLifting letStmt stmts pn destPos False
                   return (HsLetStmt (liftedDecls++ds) stmts')              
             liftToLetStmt _=mzero

             failure=idTP `adhocTP` mod
                where
                  mod (m::HsModuleP)
                   = error ( "Lifting this definition failed. "++
                           " This might be because that the definition to be lifted is defined in a class/instance declaration.")

             worker dest ds pn
                  =do let (before, parent,after)=divideDecls ds pn                                    
                          liftedDecls=definingDecls [pn] (declsIn parent) True
                          declaredPns=nub $ concatMap definedNames liftedDecls
                      (_, dd)<-hsFreeAndDeclared dest 
                      pns<-pnsNeedRenaming inscps dest parent liftedDecls declaredPns
                      if pns==[]
                        then do (parent',liftedDecls',paramAdded)<-addParamsToParentAndLiftedDecl pn dd parent liftedDecls 
                                let liftedDecls''=if paramAdded then filter isFunOrPatBind liftedDecls'
                                                                else liftedDecls'
                                 --True means the new decl will be at the same level with its parant. 
                                parent'<-moveDecl declaredPns parent  Nothing True paramAdded parent' 
                                return (before++parent'++liftedDecls''++after)
                        else askRenamingMsg pns "lifting"
                              
             doLifting  dest parent pn destPos sameLevel
                 = do let liftedDecls=definingDecls [pn] (declsIn parent) True
                          declaredPns=nub $ concatMap definedNames liftedDecls
                      pns<-pnsNeedRenaming inscps dest parent liftedDecls declaredPns
                      (_, dd)<-hsFreeAndDeclared dest 
                      if pns==[]
                        then do (parent',liftedDecls',paramAdded)<-addParamsToParentAndLiftedDecl pn dd parent liftedDecls
                                let liftedDecls''=if paramAdded then filter isFunOrPatBind liftedDecls'
                                                                else liftedDecls'
                                parent'<-moveDecl declaredPns dest destPos False  paramAdded parent'  
                                return (parent', liftedDecls'')
                        else askRenamingMsg pns "lifting"

liftedToTopLevel pn (mod@(HsModule loc name exps imps ds):: HsModuleP)  
  = if findDecl pn (declsIn ds) False 
     then let (_, parent,_) = divideDecls ds pn
              liftedDecls=definingDecls [pn] (declsIn parent) True                                    
              declaredPns  = nub $ concatMap definedNames liftedDecls
          in (True, declaredPns)
     else (False, [])

addParamsToParentAndLiftedDecl pn dd parent liftedDecls 
  =do  (ef,_)<-hsFreeAndDeclared parent
       (lf,_)<-hsFreeAndDeclared liftedDecls   
       let newParams=((nub lf)\\ (nub ef)) \\ dd  --parameters (in PName format) to be added to pn because of lifting
       if newParams/=[]  
         then if  (any isComplexPatBind liftedDecls)
                then error "This pattern binding cannot be lifted, as it uses some other local bindings!"
                else do parent'<-addParamsToParent pn newParams parent
                        liftedDecls'<-addParamsToDecls liftedDecls pn newParams
                        return (parent', liftedDecls',True)
         else return (parent,liftedDecls,False)

            
moveDecl pns dest destPos sameLevel paramsAddedOrNot
          =applyTP ((once_tdTP (failTP `adhocTP` rmInMatch
                                                 `adhocTP` rmInPat
                                                 `adhocTP` rmInLet
                                                 `adhocTP` rmInAlt
                                                 `adhocTP` rmInLetStmt)) `choiceTP` failure)
                       
           where          
            --1. The definition to be removed is a local declaration in a match
             rmInMatch (match@(HsMatch loc name pats rhs ds)::HsMatchP)
                 |findDecl pn  ds False    
                   =do ds'<-doMoving pns dest destPos sameLevel paramsAddedOrNot ds 
                       return (HsMatch loc name pats rhs ds')
             rmInMatch _ =mzero

             --2. The definition to be removed is a local declaration in a pattern binding
             rmInPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                |findDecl pn  ds  False 
                   =do ds'<-doMoving pns dest destPos sameLevel paramsAddedOrNot ds
                       return (Dec (HsPatBind  loc p rhs ds'))                      
             rmInPat _=mzero
              
             --3.The definition to be removed is a local declaration in a let expression
             rmInLet (letExp@(Exp (HsLet ds e))::HsExpP)
               |findDecl pn  ds False 
                  =do ds'<-doMoving pns dest destPos sameLevel paramsAddedOrNot  ds
                      if ds'==[] then return e
                                 else return (Exp (HsLet ds' e))
             rmInLet (letExp@(Exp (HsListComp (HsLetStmt ds stmts))))  -- e.g. [0|z=1] => [0]
               | findDecl pn  ds False 
                 =do ds'<-doMoving pns dest  destPos sameLevel paramsAddedOrNot  ds
                     if ds'==[] && isLast stmts
                         then return (Exp (HsList [fromJust (expInLast stmts)]))
                         else return (Exp (HsListComp (HsLetStmt ds' stmts)))
             rmInLet _=mzero 
             --4. The defintion to be removed is a local decl in a case alternative.
             rmInAlt (alt@(HsAlt loc p rhs ds)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
                |findDecl pn  ds False 
                =do ds'<-doMoving pns dest destPos sameLevel paramsAddedOrNot ds
                    return (HsAlt loc p rhs ds')
             rmInAlt _=mzero
            
             rmInLetStmt (letStmt@(HsLetStmt ds stmts)::(HsStmt (HsExpP) (HsPatP) [HsDeclP]))
                |findDecl pn  ds False 
               =do ds'<-doMoving pns dest destPos sameLevel paramsAddedOrNot ds
                   if ds==[]  then return stmts
                              else return (HsLetStmt ds' stmts)
             rmInLetStmt _=mzero
      
             pn=ghead "moveDecl" pns

             failure=idTP `adhocTP` mod
                where
                  mod (m::HsModuleP)
                   = error "Refactoring failed!"  --SHOULD GIVE MORE DETAILED MESSAGE


---IMPORTANT FUNCTION   
doMoving pns dest destPos sameLevel paramsAdded decls 
   = do  
        ((ts,_),_)<-get  
        --'dest' of the parent decl of the moved decl. 
        let (startPos@(_,startCol),endPos)=startEndLocIncFowNewLn ts dest
            typeSig=filter isTypeSig (definingDecls pns decls True)
        decls'<-foldM (flip rmTypeSig) decls pns
        --decls'' is the remained decls after deleting the decl defining pns
        --defToks is the tokens that have been removed. These token will be moved to another level.
        (decls'',defToks)<-rmLocalFunOrPatBind (ghead "doMoving" pns) decls'  --pns can not be empty
        --get the tokens after removing the selected declaration.
        ((toks,_),others)<-get     
        if isJust destPos  --The destination position is specified.
           then do let (destRow,destCol)=fromJust destPos  --get the destination position
                       (toks1,toks2)=break (\t->tokenPos t>=(destRow,destCol)) toks
                       origCol=tokenCol $ ghead "doMoving:then" (dropWhile isWhite defToks)
                       realDestCol=length (concatMap tokenCon (filter (\t->tokenRow t==destRow) toks1))
                       offset=realDestCol-origCol+1
                       --adjust the offset of the moved declaration
                       newDeclToks=adjustOffset offset defToks False 
                       newDeclToks'=if newDeclToks==[] || endsWithNewLn (glast "doMoving" newDeclToks)
                                       then newDeclToks
                                       else newDeclToks++[newLnToken]
                       sigToks=if typeSig==[] then []
                           else [mkToken Whitespace (0,0) 
                                 $ concatMap (\s->whiteSpaces (realDestCol-1)++s++"\n") ((lines.render.ppi) typeSig)]  
                       toks'=if paramsAdded --Type signature should be changed to reflect the parameter.
                                then (toks1++newDeclToks'++whiteSpaceTokens (0,0) realDestCol++toks2)
                                else (toks1++sigToks++newDeclToks'++whiteSpaceTokens (0,0) realDestCol++toks2)
                   put ((toks',modified),others) 
           else  --The destination position is not specified.
                do let parentToks=compressEndNewLns $ getToks (startPos,endPos) toks 
                       defToks'=dropWhile isWhite defToks
                       origCol=tokenCol $ ghead "doMoving:else" defToks'
                       realDestCol=if sameLevel then startCol else startCol+6
                       offset=realDestCol-origCol
                       --adjust the offset of the moved declaration
                       newDeclToks=adjustOffset offset defToks' False
                       --add preceding white spaces and following '\n' if necessary
                       newDeclToks'=compressPreNewLns $ whiteSpaceTokens (0,0) (realDestCol-1)++newDeclToks
                       -- add type signature if there is one
                       sigToks=if typeSig==[] then []
                               else [mkToken Whitespace (0,0)  --pretty print the type signature.
                                    $ concatMap (\s->whiteSpaces (realDestCol-1)++s++"\n") ((lines.render.ppi) typeSig)] 
                       newDeclToks''=if paramsAdded then newDeclToks'
                                          else sigToks++ newDeclToks'                                        
                       newDeclToks'''=if sameLevel 
                                        then  newDeclToks''   --Any Problem here?
                                        else (mkToken Special (0,0) (whiteSpaces (realDestCol-4) ++"where\n")):newDeclToks''
                       --handle new lines
                       toks'=if endsWithNewLn (lastNotSpaceToken parentToks) 
                                 && startsWithNewLn (firstNotSpaceToken  newDeclToks''')
                             then replaceToks toks startPos endPos (parentToks++newDeclToks''')
                             else if endsWithNewLn (lastNotSpaceToken parentToks) 
                                      || startsWithNewLn (firstNotSpaceToken  newDeclToks''')
                                    then if sameLevel  
                                           then  replaceToks toks startPos endPos (parentToks++[newLnToken]++newDeclToks''')
                                           else  replaceToks toks startPos endPos (parentToks++newDeclToks''')
                                    else if sameLevel
                                         then replaceToks toks startPos endPos (parentToks++[newLnToken,newLnToken]++newDeclToks''')
                                         else replaceToks toks startPos endPos (parentToks++[newLnToken]++newDeclToks''')
                   put ((toks',modified),others)            
        return decls''
 
----------------------------------End of Lifting-----------------------------------------

{-Refactoring : demote a function/pattern binding(simpe or complex) to the declaration where it is used.
  Descritption: if a declaration D, say, is only used by another declaration F,say, then D can be 
                demoted into the local declaration list (where clause) in F.
                So currently, D can not be demoted if more than one declaration use it. 
                
                In a multi-module context, a top-level definition can not be demoted if it is used
                by other modules. In the case that the demoted identifer is in the hiding list of
                import declaration in a client module, it should be removed from the hiding list.

 Function name:demote
 parameters: fileName--current file name.
             mod -- the scoped abstract syntax tree of the module.
             pn  -- the function/pattern name to be demoted.
       
-}

demote modName fileName (mod,tokList) pn
  =if isFunOrPatName pn mod 
    then if isTopLevelPN pn && isExplictlyExported pn mod 
          then error "This definition can not be demoted, as it is explicitly exported by the current module!"
          else do (mod',((tokList',m),_))<-doDemoting pn fileName mod tokList
                  if isTopLevelPN pn && wholeModIsExported mod  
                    then do let demotedDecls'= definingDecls [pn] (declsIn mod) True
                                declaredPns  = nub $ concatMap definedNames demotedDecls'
                            clients<-clientModsAndFiles modName
                            refactoredClients <-mapM (demotingInClientMod declaredPns) clients
                            writeRefactoredFiles $ ((fileName,m),(tokList',mod')):refactoredClients     
                    else writeRefactoredFiles [((fileName,m), (tokList',mod'))]
    else error "\nInvalid cursor position!"


--Do refactoring in the client module, that is:
--a) Check whether the identifier is used in the module body
--b) If the identifier is not used but is hided by the import declaration, then remove it from the hiding.

demotingInClientMod pns (modName, fileName)
  = do (inscps, exps, mod ,ts) <- parseSourceFile fileName
       if any (\pn->findPName pn (hsModDecls mod) || findPName pn (hsModExports mod)) pns
          then error $ "This definition can not be demoted, as it is used in the client module '"++show modName++"'!"
          else if any (\pn->findPName pn (hsModImports mod)) pns 
                  then do (mod',((ts',m),_))<-runStateT (rmItemFromImport mod pns) ((ts,unmodified),(fileName, initValue))
                          return ((fileName,m), (ts',mod'))
                  else return ((fileName,unmodified), (ts,mod)) 


doDemoting  pn fileName mod tokList
 =runStateT (applyTP ((once_tdTP (failTP `adhocTP` demoteInMod
                                         `adhocTP` demoteInMatch 
                                         `adhocTP` demoteInPat
                                         `adhocTP` demoteInLet
                                         `adhocTP` demoteInAlt
                                         `adhocTP` demoteInStmt)) `choiceTP` failure) mod)
                     ((tokList,unmodified),(fileName++".hs", initValue))
    where
       --1. demote from top level
       demoteInMod (mod@(HsModule loc name exps imps ds):: HsModuleP)  
         |findDecl pn ds False = doDemoting' mod pn
       demoteInMod _ =mzero
       
        --2. The demoted definition is a local decl in a match  
       demoteInMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
         | findDecl pn ds False = doDemoting' match pn
       demoteInMatch  _ =mzero

       --3. The demoted definition is a local decl in a pattern binding
       demoteInPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
         | findDecl pn ds False = doDemoting' pat pn           
       demoteInPat _ =mzero

       --4: The demoted definition is a local decl in a Let expression
       demoteInLet (letExp@(Exp (HsLet ds e))::HsExpP)
         | findDecl pn  ds False = doDemoting' letExp pn
       demoteInLet _=mzero
                
       --5. The demoted definition is a local decl in a case alternative.
       demoteInAlt (alt@(HsAlt loc p rhs ds)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
         |findDecl pn ds False= doDemoting'  alt pn
       demoteInAlt _=mzero
      
       --6.The demoted definition is a local decl in a Let statement.
       demoteInStmt (letStmt@(HsLetStmt ds stmts):: (HsStmt (HsExpP) (HsPatP) [HsDeclP]))
         |findDecl pn  ds False = doDemoting' letStmt pn
       demoteInStmt _=mzero

       failure=idTP `adhocTP` mod
             where
               mod (m::HsModuleP)
                = error "Refactoring failed!"   --SHOULD GIVE MORE DETAILED ERROR MESSAGE
     
{- doDemoting' :(MonadPlus m)=>PName->[HsDeclP]->m [HsDeclP] 
   parameters:  t -declaration or expression  where pn is define.
                pn -- the function/pattern name to be demoted in PName format    
            
-}
doDemoting' t pn 
 = let origDecls=declsIn t
       demotedDecls'=definingDecls [pn] origDecls True
       declaredPns=nub $ concatMap definedNames demotedDecls'
       demotedDecls=definingDecls declaredPns origDecls True
   in if not (usedByRhs t declaredPns)
       then do -- find how many matches/pattern bindings (except the binding defining pn) use 'pn' 
              uselist<-uses declaredPns (declsIn t)
              case  length uselist  of
                  0 ->error "\n Nowhere to demote this function!\n"
                  1 -> --This function is only used by one friend function 
                      do (f,d)<-hsFreeAndDeclared demotedDecls 
                          -- remove demoted declarations
                         --Without updating the token stream.
                         let ds=foldl (flip removeTypeSig) (declsIn t\\demotedDecls) declaredPns  
                         --get those varaibles declared at where the demotedDecls will be demoted to
                         dl  <-mapM (flip declaredNamesInTargetPlace ds) declaredPns
                         --make sure free variable in 'f' do not clash with variables in 'dl', 
                         --otherwise do renaming. 
                         let clashedNames=filter (\x-> elem (pNtoName x) (map pNtoName f)) $ (nub.concat) dl
                         --rename clashed names to new names created automatically,update TOKEN STREAM as well.
                         if clashedNames/=[] 
                            then error ("The identifier(s):" ++ showPns pNtoQuotedName clashedNames ++
                                       ", declared in where the definition will be demoted to, will cause name clash/capture"
                                       ++" after demoting, please do renaming first!")  
                                 --ds'<-foldM (flip (autoRenameLocalName True)) ds clashedNames
                            else  --duplicate demoted declarations to the right place.
                                 do ds''<-duplicateDecls declaredPns demotedDecls origDecls ds
                                    return (replaceDecls t ds'')
                  _ ->error "\nThis function/pattern binding is used by more than one friend bindings\n"
                       
      else error "This function can not be demoted as it is used in current level!\n"
             
    where
          ---find how many matches/pattern bindings use  'pn'------- 
          uses pns 
               = applyTU (stop_tdTU (failTU `adhocTU` usedInMatch
                                            `adhocTU` usedInPat))                           
                where
                  usedInMatch (match@(HsMatch _ (PNT pname _ _) _ _ _)::HsMatchP)
                     | isNothing (find (==pname) pns) && any  (flip findPName match) pns
                     =return ["Once"]
                  usedInMatch _ =mzero

                  usedInPat (pat@(Dec (HsPatBind _ p _ _)):: HsDeclP)
                    | pNamesInPat p `intersect` pns ==[]  && any  (flip findPName pat) pns
                    =return ["Once"]
                  usedInPat  _=mzero 

          -- duplicate demotedDecls to the right place (the outer most level where it is used).
          duplicateDecls  pns demotedDecls decls
             =applyTP (once_tdTP (failTP `adhocTP` dupInMatch 
                                         `adhocTP` dupInPat)) 
               where 
                 dupInMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
                   | any (flip findPName match) pns 
                   =  {-If do not fold parameters:  (I think it's better to seperate folding parameters and demoting) -}
                     do --doMoving1 pns match decls False
                        --return (HsMatch loc1 name pats rhs (ds++demotedDecls))
                        foldParams pns match decls demotedDecls
                 dupInMatch _ =mzero
     
                 dupInPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                    |any (flip findPName pat) pns
                   =  do doMoving1 pns pat decls False
                         return (Dec (HsPatBind loc p rhs (ds++demotedDecls)))
                 dupInPat _ =mzero  
                 
          ---------------------------------------------------------------------
          declaredNamesInTargetPlace :: (Term t, MonadPlus m)=>PName->t->m [PName]
          declaredNamesInTargetPlace pn=applyTU (stop_tdTU (failTU 
                                                    `adhocTU` inMatch  
                                                    `adhocTU` inPat))                            
               where
                 inMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
                    | findPName pn rhs
                     =(return.snd)=<<freeAndDeclaredVbls match 
                 inMatch _ =mzero

                 inPat (pat@(Dec (HsPatBind loc p rhs ds)):: HsDeclP)
                    |findPName pn rhs
                     =(return.snd)=<<freeAndDeclaredVbls pat 
                 inPat _=mzero  

 ---IMPORTANT FUNCTION. this function is only used in demoting.
--The function name should be renamed later.
doMoving1 pns dest decls paramsFolded
   = do ((ts,_),_)<-get
        let (startPos@(_,startCol),endPos)=startEndLocIncFowComment ts dest           
            typeSig=filter isTypeSig (definingDecls pns decls True )
        decls'<-foldM (flip rmTypeSig) decls pns 
        (decls'',defToks)<-rmLocalFunOrPatBind (ghead "doMoving1"  pns) decls'   ---PROBLEM HERE?
        ((toks,_),others)<-get   
        let parentToks=compressEndNewLns $ getToks (startPos,endPos) toks  
            origCol=tokenCol $ ghead "doMoving1" (dropWhile isWhite defToks)
            localDecls=declsIn dest
            realDestCol=if localDecls==[] then startCol+6
                                          else (snd.fst) (startEndLoc toks (ghead "doMoving" localDecls))  --no problem with head
            offset=realDestCol-origCol
            typeSigLines=(nub ((lines.render.ppi) typeSig)) \\[" "] --remove the empty lines between type sig.
            sigToks=if typeSig==[] then []
                        else [mkToken Varid (0,0) 
                             $ concatMap (\s->whiteSpaces (realDestCol-1)++s++"\n") (init typeSigLines) ++
                               whiteSpaces (realDestCol-1)++glast "doMoving1" typeSigLines]  --no problem with 'last'

            newDeclToks=adjustOffset offset (compressPreNewLns defToks) True
            newDeclToks'=newDeclToks --whiteSpaceTokens (0,0) (realDestCol-1)++newDeclToks
            newDeclToks''=if paramsFolded then newDeclToks'
                                          else if startsWithEmptyLn newDeclToks' 
                                                  then sigToks++newDeclToks'           
                                                   else sigToks++[newLnToken]++newDeclToks'
            newDeclToks'''=
                    if localDecls/=[]
                       then newDeclToks''
                       else if startsWithEmptyLn newDeclToks'' 
                              then (mkToken Special (0,0) (whiteSpaces (startCol-1)++"  where")):newDeclToks''
                              else (mkToken Special (0,0) (whiteSpaces (startCol-1)++"  where\n")):newDeclToks''
            toks'=if  endsWithNewLn (lastNotSpaceToken  parentToks) || startsWithEmptyLn newDeclToks'''
                   then replaceToks toks startPos endPos (parentToks++newDeclToks''')
                   else replaceToks toks startPos endPos (parentToks++[newLnToken]++newDeclToks''')     
        put ((toks',modified),others)



class (Term t) =>UsedByRhs t where
  
    usedByRhs:: t->[PName]->Bool

instance UsedByRhs HsExpP where
    usedByRhs (Exp (HsLet ds e)) pns = or $ map (flip findPName e) pns

instance UsedByRhs HsAltP where
    usedByRhs (HsAlt _ _ rhs _) pns  =or $ map (flip findPName rhs) pns 

instance UsedByRhs HsStmtP where
    usedByRhs (HsLetStmt _ stmt) pns =or $ map (flip findPName stmt) pns  

instance UsedByRhs HsMatchP where   
    usedByRhs (HsMatch loc1 fun pats rhs ds) pns =or $ map (flip findPName rhs) pns   
           
instance UsedByRhs  HsDeclP where
    usedByRhs (Dec (HsPatBind loc p rhs ds)) pns =or $ map (flip findPName rhs) pns               
    usedByRhs _ pn=False

instance UsedByRhs HsModuleP where
    usedByRhs mod pns=False


{- foldParams:remove parameters in the demotedDecls if possible
   parameters: pn -- the function/pattern name to be demoted in PName format
               match--where the demotedDecls will be demoted to
               demotedDecls -- the declarations to be demoted.
   example:
    module Test where        demote 'sq'       module Test where
    sumSquares x y               ===>          sumSquares x y =(sq 0) + (sq y) 
      = sq x 0+ sq x y                               where sq y=x ^ y
    sq x y=x^y
-}
--PROBLEM: TYPE SIGNATURE SHOULD BE CHANGED.

foldParams pns (match@(HsMatch loc1 name pats rhs ds)::HsMatchP) decls demotedDecls 
     =do let matches=concatMap matchesInDecls demotedDecls
             pn=ghead "foldParams" pns    --pns /=[]
         params<-allParams pn rhs []
         if (length.nub.map length) params==1                  -- have same number of param 
             -- && (all (\x->length x ==1) (map nub params'))  -- have exactly same parameters
              && ((length matches)==1)      -- only one 'match' in the demoted declaration             
           then do let patsInDemotedDecls=(patsInMatch.(ghead "foldParams")) matches
                       subst=mkSubst patsInDemotedDecls params
                       fstSubst=map fst subst
                       sndSubst=map snd subst
                   rhs'<-rmParamsInParent pn sndSubst rhs
                   ls<-mapM hsFreeAndDeclared sndSubst
                   -- newNames contains the newly introduced names to the demoted decls---
                   let newNames=(map pNtoName (concatMap fst ls)) \\ (map pNtoName fstSubst)
                   --There may be name clashing because of introducing new names.
                   clashedNames<-getClashedNames fstSubst newNames (ghead "foldParams" matches)
                   --auotmatic renaming 
                   demotedDecls'<-foldM (flip (autoRenameLocalName True)) demotedDecls clashedNames
                   demotedDecls''<- foldM replaceExpWithUpdToks demotedDecls' subst 
                   --remove substituted parameters in demoted declarations
                   demotedDecls'''<-mkDefaultLocsUnique =<< rmParamsInDemotedDecls fstSubst demotedDecls''  
                   doMoving1 [pn] match decls True
                   return (HsMatch loc1 name pats rhs' (ds++(filter (not.isTypeSig) demotedDecls''')))
           else do doMoving1 pns match decls False
                   return (HsMatch loc1 name pats rhs (ds++demotedDecls))  -- no parameter folding 

    where  
      ------Get all of the paramaters supplied to pn --------------------------- 
            {- eg. sumSquares x1 y1 x2 y2 = rt x1 y1 + rt x2 y2
                   rt x y = x+y
              demote 'rt' to 'sumSquares',
              'allParams pn rhs []'  returns [[x1,x2],[y1,y2]]
                where pn is 'rt' and  rhs is 'rt x1 y1 + rt x2 y2'             
           -}
       allParams pn rhs initial  -- pn: demoted function/pattern name.
        =do (fileName,_)<-nextValue
            p<-getOneParam fileName pn rhs
            --putStrLn (show p)
            if p/=[] then do rhs'<-rmOneParam pn rhs
                             allParams pn rhs' (initial++[p])           
                     else return initial
        where
           getOneParam fileName pn 
              =applyTU (stop_tdTU (failTU `adhocTU` worker))
                where
                  worker (Exp (HsApp e1 e2))
                   |(expToPName e1==pn) =return (rmLocInfo1 fileName [e2])
                  worker _ =mzero
           rmOneParam pn
              =applyTP (stop_tdTP (failTP `adhocTP` worker))
                where
                  worker (Exp (HsApp e1 e2 ))
                    |expToPName e1==pn =return e1
                  worker _ =mzero
      
       -----------remove parameters in demotedDecls-------------------------------
       rmParamsInDemotedDecls ps 
         =applyTP (once_tdTP (failTP `adhocTP` worker))  
            where worker ((HsMatch loc1 name pats rhs ds)::HsMatchP)
                    = do ((toks,_),others)<-get   
                         let (startPos,endPos)=startEndLoc toks pats 
                             pats'=filter (\x->not ((patToPName x /=defaultPN) && 
                                          elem (patToPName x) ps)) pats
                             newPatToks=concatMap (\p->getToks (startEndLoc toks p) toks)  pats'
                             toks'=replaceToks toks startPos endPos newPatToks
                         put ((toks',modified),others)
                         return (HsMatch loc1 name pats' rhs ds)

    
       ----------remove parameters in the parent functions' rhs-------------------
       --Attention: PNT i1 _ _==PNT i2 _ _ = i1 =i2 
       rmParamsInParent  pn es
         =applyTP (full_buTP (idTP `adhocTP` worker))
            where worker exp@(Exp (HsApp e1 e2))
                   | findPName pn e1 && elem e2 es
                      =do ((toks,_),others)<-get   
                          let (startPos,endPos)=startEndLoc toks exp
                              toks'=replaceToks toks startPos endPos $ getToks (startEndLoc toks e1) toks
                          put ((toks',modified),others)
                          return e1
                  worker (exp@(Exp (HsParen e1)))
                    |pn==expToPName e1
                       =do ((toks,_),others)<-get   
                           let (startPos,endPos)=startEndLoc toks exp
                               toks'=replaceToks toks startPos endPos $ getToks (startEndLoc toks e1) toks
                           put ((toks',modified),others)
                           return e1 
                  worker x =return x 
            
       getClashedNames oldNames newNames (match::HsMatchP)
         = do  (f,d)<-freeAndDeclaredVbls match 
               ds'<-mapM (flip hsVisibleVbls match) oldNames
               -- return clashed names
               return (filter (\x->elem (pNtoName x) newNames)  --Attention: nub
                                   ( nub (d `union` (nub.concat) ds')))
       ----- make Substitions between formal and actual parameters.-----------------
       mkSubst pats params
           = catMaybes (zipWith (\x y ->if (patToPName x/=defaultPN) && (length (nub y)==1)
                            then Just (patToPName x,(ghead "mkSubst") y)
                            else Nothing) pats params)  
     

--substitute an old expression by new expression
replaceExpWithUpdToks  decls subst
   = applyTP (full_buTP (idTP `adhocTP` worker)) decls
         where worker (e::HsExpP)
                 |(expToPName e/=defaultPN) &&  (expToPName e)==(fst subst)
                     = do ((toks,_),others)<-get   
                          let (startPos,endPos)=startEndLoc toks e
                              newExpToks=mkTokensNoFstLnIndent Varid startPos ((lines.render.ppi) (snd subst))
                              toks'=replaceToks toks startPos endPos newExpToks 
                          put ((toks',modified),others)
                          return (snd subst)
               worker x=return x 



