

module RefacRmDef(removeDef,rmTypeSig, removeInTopLevel, rmLocalFunOrPatBind) where

import PosSyntax
import Maybe
import TypedIds
import UniqueNames
import List 
import TiPNT 
import RefacUtils 

{-This refactoring removes a user selected function binding or a pattern binding if it is not used elsewhere
  When a function/pattern is removed, it's corresponding type signature should be removed as well. In the case
  that the type signature is like this a,b:: blah balh, only the name is removed from the type signature.
  A function binding  can be removed if the function name is not used elsewhere; A pattern binding can be 
  removed only if none of the names defined in this pattern binding is used elsewhere.
-}

removeDef args
 = do let fileName = ghead "filename" args 
          --fileName'= moduleName fileName
          --modName  = Module fileName'  
          row      = read (args!!1)::Int
          col      = read (args!!2)::Int
      modName <- moduleName fileName 
      (inscps, _, mod, tokList)<-parseSourceFile fileName         
      let pn = locToPName fileName row col mod                                            
      if (pn /= defaultPN) 
       then if isTopLevelPN pn  
             then do let pns=pnsToBeRemoved pn mod 
                     if any (flip isExplictlyExported mod) pns
                       then error "This definition can not be removed, as it is explicitly exported by this module!"
                       else do (mod',((tokList',modified),_))<-doRemoving pn fileName mod tokList
                               if isTopLevelPN pn 
                                then do clients<-clientModsAndFiles modName
                                        refactoredClients <- mapM (refactorInClientMod pns) clients
                                        writeRefactoredFiles $ ((fileName,modified),(tokList',mod')):refactoredClients
                                else writeRefactoredFiles [((fileName,modified), (tokList',mod'))]
             else do (mod',((tokList',modified),_))<-doRemoving pn fileName mod tokList
                     writeRefactoredFiles [((fileName,modified), (tokList',mod'))]
       else error "\nInvalid cursor position!" 
  
doRemoving pn fileName mod tokList 
    =  runStateT (applyTP ((once_tdTP (failTP `adhocTP` rmInMod
                                              `adhocTP` rmInMatch
                                              `adhocTP` rmInPat
                                              `adhocTP` rmInLet
                                              `adhocTP` rmInAlt
                                              `adhocTP` rmInLetStmt)) `choiceTP` failure) mod) ((tokList,unmodified),(fileName,initValue))
                       
           where          
             --1. The definition to be removed is one of the module's top level declarations.
             rmInMod (mod@(HsModule loc name exps imps ds):: HsModuleP)  
                | canBeRemoved pn mod   
                  =do ds'<-removeInTopLevel pn ds
                      return (HsModule loc name exps imps ds') 
             rmInMod _ =mzero

             --2. The definition to be removed is a local declaration in a match
             rmInMatch (match@(HsMatch loc name pats rhs ds)::HsMatchP)
                 | canBeRemoved pn match
                   =do ds'<-removeInLocal pn ds
                       return (HsMatch loc name pats rhs ds')
             rmInMatch _ =mzero

             --3. The definition to be removed is a local declaration in a pattern binding
             rmInPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                | canBeRemoved pn pat
                   =do ds'<-removeInLocal pn ds
                       return (Dec (HsPatBind  loc p rhs ds'))                      
             rmInPat _=mzero
              
             --4.The definition to be removed is a local declaration in a let expression
             rmInLet (letExp@(Exp (HsLet ds e))::HsExpP)
               | canBeRemoved pn letExp
                  =do ds'<-removeInLocal pn ds
                      if ds'==[] then return e
                                 else return (Exp (HsLet ds' e))
             rmInLet (letExp@(Exp (HsListComp (HsLetStmt ds stmts))))  -- e.g. [0|z=1] => [0]
               | canBeRemoved pn letExp
                 =do ds'<-removeInLocal pn ds
                     if ds'/=[] 
                         then return (Exp (HsListComp (HsLetStmt ds' stmts)))
                         else if isLast stmts
                                then return (Exp (HsList [fromJust (expInLast stmts)]))
                                else return (Exp (HsListComp stmts))
             rmInLet _=mzero 
       
             --5. The defintion to be removed is a local decl in a case alternative.
             rmInAlt (alt@(HsAlt loc p rhs ds)::(HsAlt (HsExpP) (HsPatP) [HsDeclP]))
                |canBeRemoved pn alt
                =do ds'<-removeInLocal pn ds
                    return (HsAlt loc p rhs ds')
             rmInAlt _=mzero
            
             --6. The definition to be removed is a local decl in a let statement.
             rmInLetStmt (letStmt@(HsLetStmt ds stmts)::(HsStmt (HsExpP) (HsPatP) [HsDeclP]))
                |canBeRemoved pn letStmt
               =do ds'<-removeInLocal pn ds
                   if ds'==[]  then return stmts
                              else return (HsLetStmt ds' stmts)
             rmInLetStmt _=mzero
      
             failure=idTP `adhocTP` mod
              where
               mod (m::HsModuleP)
                = error ("Refactoring Failed! Possible reasons: a) selected identifier is not a function/pattern name defined in this module;"++
                          " b) the identifer is used elsewhere;") 
  
             canBeRemoved pn t 
               =let decls=declsIn t
                    decl=definingDecls [pn] decls False
                    pnames=concatMap definedNames decl
                in  decl/=[] && all (not.flip findPName (replaceDecls t (decls \\ decl))) pnames

  
pnsToBeRemoved pn t
 = let decls=declsIn t
       decl=definingDecls [pn] decls False
   in concatMap definedNames decl         

removeInLocal pn decls
   =do (decls',_)<-rmLocalFunOrPatBind pn =<< rmTypeSig pn decls
       return decls'

removeInTopLevel pn decls
   =rmTopLevelFunOrPatBind pn =<< rmTypeSig pn decls 

--IMPORTANT FUNCTION
--Remove a the type signature which defines pn's type from the declaration list.
rmTypeSig pn decls  
 =let (decls1,decls2)=break  (definesType pn) decls
  in if decls2/=[] then    --There is a type signature for pn
       do ((toks,_),others)<-get   
          let (toks',decls')=
               let sig@(Dec (HsTypeSig loc is c tp))=ghead "rmTypeSig" decls2  -- as decls2/=[], no problem with head 
                   (startPos,endPos)=startEndLoc toks sig 
               in if length is>1   --This type signature also defines the type of  variables other than pn
                     then let newSig=(Dec (HsTypeSig loc (filter (\x-> (pNTtoPN x)/=pn) is) c tp))
                             --pretty print the type signature, and put it into a new token 
                             --attention:The token type can not be WhiteSpace 
                              t=mkTokensNoFstLnIndent Varid startPos  (astToSource newSig)
                          in (replaceToks toks startPos endPos t,(decls1++[newSig]++tail decls2))
                     else  ((deleteToks toks startPos endPos),(decls1++tail decls2)) 
          put ((toks',modified),others)
          return decls' 
      else return decls  --No type signature for pn

--IMPORTANT FUNCTION
--Remove the local declaration defining pn from decls(which is a list of declarations).

{- The difference between removing a top level declaration and a local declaration is:
   if the local declaration to be removed is the only declaration in current declaration list, 
   then the 'where'/ 'let'/'in' enclosing this declaration should also be removed.
   Whereas, when a only top level decl is removed, the 'where' can not be removed.
-}
rmLocalFunOrPatBind pn decls  
  =do ((toks,_),others)<-get   
      let (decls1,decls2)=break (defines pn) decls
          decl=ghead "rmLocalFunOrPatBind:decl" decls2   --the declaration to be removed, decls2 won't be empty.
          (startPos,endPos)=startEndLoc toks decl   --startEndLoc toks decl  
          (startPos',endPos')=startEndLocIncComments1 toks decl       
          --(startPos',endPos')=startEndLocIncFowComment toks decl 
          toks'=if length decls==1 --only one decl, which means the accompaning 'where','let' or'in' should be removed too.
                   then let (toks1,toks2)=break (\t->tokenPos t>=startPos) toks --devide the token stream.
                              --get the  'where' or 'let' token
                            rvToks1=dropWhile (not.isWhereOrLet) (reverse toks1)
                            --There must be a 'where' or 'let', so rvToks1 can not be empty.
                            whereOrLet=ghead "rmLocalFunPatBind:whereOrLet" rvToks1 
                            --drop the 'where' 'or 'let' token
                            toks1'=takeWhile (\t->tokenPos t<tokenPos whereOrLet) toks1
                            --remove the declaration from the token stream.
                            toks2'=dropWhile (\t->tokenPos t<=endPos') toks2
                            --get the remained tokens after the removed declaration.
                            remainedToks=dropWhile isWhite toks2'
                        in if remainedToks==[] 
                             then --the removed declaration is the last decl in the file.
                                  (compressEndNewLns toks1'++ compressPreNewLns toks2') 
                             else if --remainedToks/=[], so no problem with head.
                                    isIn (ghead "rmLocalFunPatBind:isIn"  remainedToks)  
                                         || isComma (ghead "rmLocalFunPatBind:isComma" remainedToks) 
                                        --There is a 'In' after the removed declaration.
                                   then deleteToks toks (tokenPos whereOrLet)
                                                $ tokenPos (ghead "rmLocalFunPatBind:tokenPos"  remainedToks) 
                                        --delete the decl and adjust the layout                      
                                   else if isCloseSquareBracket (ghead "rmLocalFunPatBind:isCloseSquareBracker" remainedToks) &&
                                           (isBar.(ghead "rmLocalFunPatBind:isBar")) (dropWhile isWhite (tail rvToks1))
                                         then deleteToks toks (tokenPos((ghead "rmLocalFunPatBind") 
                                                        (dropWhile isWhite (tail rvToks1)))) endPos'
                                         else deleteToks toks (tokenPos whereOrLet) endPos' 
                        --there are more than one decls
                   else  deleteToks toks startPos' endPos'
      put ((toks',modified),others)  --Change the above endPos' to endPos will not delete the following comments.
      --This function also returns the removed token stream. this can be used when moving a definition.
      return (decls1++tail decls2, getToks (startPos',endPos') toks)


--Remove a function/pattern declaration from top level.
rmTopLevelFunOrPatBind  pn  decls  
  =do ((toks,_),others)<-get   
      let (decls1,decls2)=break (defines pn) decls
          decl=ghead "rmTopLevelFunOrPatBind"  decls2  --decls2/=[], no problem with 'head'.
          (startLoc, endLoc)=startEndLocIncComments toks decl      
          toks'=deleteToks toks startLoc endLoc
      put ((toks',modified),others)
      return (decls1++tail decls2) 


refactorInClientMod pns (modName@(Module modName'), fileName)
  = do (inscps, exps, mod ,ts) <-parseSourceFile fileName
       if any (flip findPName (hsModDecls mod)) pns || any (flip findPName (hsModExports mod)) pns
          then error $ "This definition can not be removed, as it is used by the client module '"++modName'++"'!"
          else do let pnsToBeHided= filter (flip findPName (hsModImports mod)) pns 
                  if pnsToBeHided/=[]  
                     then do (mod', ((ts',m),_))<-runStateT (rmItemFromImport mod pnsToBeHided)
                                                       ((ts,unmodified),(fileName, initValue))
                             return ((fileName,m), (ts',mod'))
                     else return ((fileName,unmodified), (ts,mod)) 


--MOVE THIS FUNCTION TO "RefacLoc.hs" *****
--get the start&end location of t in the token stream, including the preceding and following
--comments.

startEndLocIncComments1 toks t
  =let (startLoc,endLoc)=startEndLoc toks t    
       toks1=takeWhile (\t->tokenPos t<startLoc) toks
       --get all the white tokens before the start location
       groupedToks=groupBy (\x y->tokenRow x==tokenRow y) (reverse toks1)
       emptyLns=takeWhile (all (\t->isWhiteSpace t || isNewLn t )) groupedToks       
       lastComment=if length emptyLns <=1 
                     then takeWhile (all isWhite) $ takeWhile (any isComment) $ dropWhile 
                                                (all (\t->isWhiteSpace t || isNewLn t)) groupedToks    
                     else []
       toks1'=reverse $ concat (emptyLns ++ lastComment)
       startLoc'=if toks1'==[] then startLoc    --there is no comment before t
                               else tokenPos (ghead "startEndLocIncComments"  toks1') --there is a comment before t
       toks2=dropWhile (\t->tokenPos t<=endLoc) toks
       (toks21,tok22)=break (\t->tokenRow t/=fst endLoc) toks2
   in if toks21==[] 
        then (startLoc',endLoc)
        else if all (\t->isWhite t || endsWithNewLn t) toks21 --get the following white tokens in the same
                                                              --line of the last token of t
                 then (startLoc', tokenPos (last toks21))
                 else (startLoc', endLoc) 

